////////////////////////////////////////////////////////////////////////
// Class:       Calibration
// Module Type: analyzer
// File:        Calibration_module.cc
//
// Generated at Wed Nov 11 14:47:12 2015 by Andrew S. Hirsch using artmod
// from cetpkgsupport v1_08_07.
////////////////////////////////////////////////////////////////////////

#include <iostream>
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Utilities/InputTag.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include <TH2I.h>
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/HitSet.h"
#include "NoiseHitMap.h"
#include "SharedNoiseMap.h"



// This anlyzer's goal is to collect data from events into calibration histograms.
// It will collect EventsForCalibration before analizing this data and store for further usage.


namespace novaddt {
  class Calibration;
}


class novaddt::Calibration : public art::EDAnalyzer {
public:
  explicit Calibration(fhicl::ParameterSet const & p);
  // The destructor generated by the compiler is fine for classes
  // without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  Calibration(Calibration const &) = delete;
  Calibration(Calibration &&) = delete;
  Calibration & operator = (Calibration const &) = delete;
  Calibration & operator = (Calibration &&) = delete;

  // Required functions.
  void analyze(art::Event const & e) override;
  
  // Selected optional functions.
  void beginJob() override;
  void endJob() override;

private:
    void accumulate(art::Event const& e);
    void recalibrate();
    TH2I* MakeNoiseHisto(const novaddt::calib::FDHitMap &m);
novaddt::HitSet fetchAllTrackHits(const art::Event & event);

private:
  // Declare member data here.
  unsigned _EventCounter = 0;
  unsigned _EventsToCollect = 50;  //number of events to accumulate for calibration
  std::vector<std::string> _hits_id;   //[label,instance] of hits
  std::vector<std::string> _tracks_id; //[label, instance] of tracks
  SharedFDNoiseMap _SharedNoiseMap;
};


novaddt::Calibration::Calibration(fhicl::ParameterSet const & p):
  EDAnalyzer(p),
  _EventsToCollect(p.get<unsigned>("EventsToCollect")),
  _hits_id(p.get<std::vector<std::string>>("hits_id")),
  _tracks_id(p.get<std::vector<std::string>>("tracks_id")),
  _SharedNoiseMap("calibShmem","FDNoiseHitMap")
{}

void novaddt::Calibration::analyze(art::Event const & e)
{
    ++_EventCounter;
     // Hitlist from event (input)
    art::Handle<novaddt::HitList> hitsHandle;
    e.getByLabel(_hits_id[0],_hits_id[1], hitsHandle);

    //Create a set of all hits in tracks
    novaddt::HitSet allTrackHits=fetchAllTrackHits(e);
    novaddt::HitSet freeHits(*hitsHandle);
    freeHits.substract(allTrackHits);

    //get access to the shared hit map
    auto& BufManager = _SharedNoiseMap.getData();
    //increment event counter
    std::cout<<"Counter="<<BufManager.getCounter()<<std::endl;
    BufManager.addCounter();
    //obtain sharable lock, to indicate we're currently R\W data in map
    auto lock = BufManager.sharable_lock();
    auto &NoiseMap = BufManager.getWBuffer();
    //fill the hits map
    for(const auto &hit: freeHits)
        NoiseMap[hit]+=1;
    
    //now try reading the previous map
    if(BufManager.getCounter()==1){
        MakeNoiseHisto(BufManager.getRBuffer());    
    }

}

novaddt::HitSet novaddt::Calibration::fetchAllTrackHits(const art::Event & event){
    //Create a set of all hits, that belong to tracks
    novaddt::HitSet allTrackHits;
    //Fill this set from tracks
    art::Handle<std::vector<novaddt::HitList>> trackHitLists;
    event.getByLabel(_tracks_id[0],_tracks_id[1],trackHitLists);
    for(const auto& trackHits: *trackHitLists)
        allTrackHits.add(trackHits);
    return allTrackHits;
}


TH2I* novaddt::Calibration::MakeNoiseHisto(const novaddt::calib::FDHitMap& m){
   art::ServiceHandle<art::TFileService> tfs;
   auto hist=tfs->make<TH2I>(Form("h_%d",_EventCounter),
                        Form("Noise, event #%d",_EventCounter),
                        m.Nplanes(),0,m.Nplanes(),
                        m.Ncells(),0,m.Ncells());
   for(size_t nc=0;nc<m.Ncells();++nc)
     for(size_t np=0;np<m.Nplanes();++np){
        hist->SetBinContent(np,nc,m(novaddt::Plane(np),novaddt::Cell(nc)));
     }
   std::cout<<"Made hist with total I="<<hist->Integral()<<std::endl;
   return hist;
}

void novaddt::Calibration::beginJob()
{
    std::cout<<"---- Calibration analyzer options -----"<<std::endl;
    std::cout<<"\tEventsToCollect = "<<_EventsToCollect <<std::endl;
    std::cout<<"\tInput hits label   = ["<<_hits_id[0]  <<" : "<<_hits_id[1]  <<"]"<<std::endl;
    std::cout<<"\tInput tracks label = ["<<_tracks_id[0]<<" : "<<_tracks_id[1]<<"]"<<std::endl;
    
}

void novaddt::Calibration::endJob()
{
  // Implementation of optional member function here.
}

DEFINE_ART_MODULE(novaddt::Calibration)
